/**********************************************************************************
 * NL nl.c
 * =================
 * versione    03
 * autore      Calzoni Pietro aka Calzo <calzog@gmail.com>
 *               LUGMan - Linux Users Group Mantova
 *               www.lugman.org
 * licenza     GPL
 * Kernel      2.6
 * Kernel test 2.6.13.1
 * System      SalckWare 10.2
 * compilers   gcc-3.3.6
 * --------------------------------------------------------------------------------
 * Dalla versione 02 si  visto come sia possibile ottenere un messaggio dallo
 * user-space, prelevandone le informazioni.
 * In questo modulo invece si vuole fare in modo che anche il kernel invii un
 * messaggio unicast allo user-space identificnado il mittente dal suo processo
 * --------------------------------------------------------------------------------
 * IMPEMENTAZIONE:
 * l'implementazione  relativa principalmente alla callback input() associata al
 * socket. L'accesso ai dati provenienti da user-space viene fatto con:
 * - sk_buff: definito in skbuff.h. Ha una serie infinita di campi anche a union
 *  e tra gli altri permette l'accesso diretto al buffer del socket tramite il
 *  campo data. Questo campo contiene anche l'header nlmsghdr.
 * - nlmsghdr: NETLINK Message Header definito in linux/netlink.h. Tra i vari campi
 *  continene anche nlmsg_pid ossia il pid del progesso mittente del messaggio.
 * Occorre decidere come "pilotare" l'invio del messaggio da parte del kernel. Ci
 * lo si pu fare in molti modi (proc, device driver, ecc), ma la via pi semplice
 * forse  inviare un messaggio con la richiesta di invio del messaggio.
 * Per ragioni che saranno chiare pi avanti introduco una struttua per formattare
 * il payload. Vedere `nl.h' per la strttura.
 * L'invio di un messaggio da parte del kernel viene effettuato con la chiamata:
 *     netlink_unicast(struct sock, struct sk_buff, u32, int);
 * Come si nota la strtuttura necessaria  sempre sk_buff a cui si accede mediate
 * la macro NETLINK_CB() definita in linux/netlink.h
 * --------------------------------------------------------------------------------
 * TEST:
 * Il test  condotto con nl_send per la versione 03 del modulo. Riferirisi alla
 * documentazione interna al file nl_send.c per vedere le varie modalit di test.
 *********************************************************************************/
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
//--- NetLink ---
#include <linux/socket.h>  // gestione dei socket e strutture relative
#include <linux/netlink.h> // implementato in net/netlink/af_netlink.c (2.6.13.1)
#include <net/sock.h>      // implementa la struttura sock. senza questo  un casino referenziare i puntatori
//--- nl.h ---
#include "nl.h"            // comprende macro e variabili necessarie al kernel quanto allo user space

MODULE_LICENSE("GPL");
MODULE_VERSION("0.3");
MODULE_AUTHOR("Calzo");
MODULE_DESCRIPTION("NL creates a NetLink socket and test the comunication between user and kernel space");

static struct sock *nl_sk = NULL;

/** _change_str_case non fa altro che cambiare da minuscola a maiuscola la
 * stringa passatagli
 * */
void nl_change_str_case(char *str, int maxstr_len)
{
   int i;
   
   for(i=0; str[i] && i<maxstr_len; i++)
     {
	if(str[i]>='A' && str[i]<='Z')
	  str[i]+=0x20;
	else
	  if(str[i]>='a' && str[i]<='z')
	    str[i]-=0x20;
     }
}


/** _input  una callback che deve essere definita e passata alla funzione di
 * creazione del ntlink socket `netlink_kernel_create()'.
 * Questa funzione invia dal kernel allo user-space un messaggio a fronte di un
 * altro messaggio proveniente dallo user-space. Questo serve solo per triggherare
 * l'evento di invio da parte del kernel ed  una scelta arbitraria.
 * --------------------------------------------------------------------------------
 *  NOTA: se il manage del socket  abbastanza leggero, la funzione seguente pu
 * occuparsene direttamente, altrimenti  bene eseguire un wake_up_ in modo da
 * sbloccare l'esecuzione di un task sospeso appositamente.
 *  NOTA: MSG_DONTWAIT permette di rendere la chiamata non-interruttiva. Credo che
 * questo provochi la perdita del messaggio se nessun socket  in ascolto, ma non
 * ne sono sicuro.
 * */
void nl_input(struct sock *sk, int len)
{
   struct sk_buff *buf = NULL;
   struct nlmsghdr *nlh;
   int err;
   struct nl_payload_t *payload;
   
   printk(KERN_ALERT "Messaggio ricevuto da user-space\n");
   
   // Prelevo dalla coda il buffer. Per gestire tutta la coda dei messaggi credo sia mglio
   // usare skb_dequeue(&sk->receive_queue)
   buf = skb_recv_datagram(sk, 0, 0, &err);
   nlh = (struct nlmsghdr *)buf->data;                // prelevo il la struttura del messaggio compreso l'header NETLINK
   payload = (struct nl_payload_t *)NLMSG_DATA(nlh);  // prelevo il payload, ossia i dati effettivi
   printk(KERN_ALERT "  sender pid: %d > %s\n", 
	  nlh->nlmsg_pid, payload->str);
   
   // A questo punto conosco il mittente (o il presunto tale). 
   // Reinvio un messaggio unicast al pid
   if(payload->sender_pid == 0)    // Se viene richiesto di inviare il messaggio AL
     return;                       //  kernel, ritorno per evitare casini ;)
   // Rielaboro il Payload tanto per fare vedere qualche cosa
   nl_change_str_case(payload->str, MAX_PAYLOAD_CHAR);
   // Setto i parametri per l'invio
   NETLINK_CB(buf).groups = 0;     // Nessun gruppo per multicast
   NETLINK_CB(buf).pid = 0;        // Il processo mittente  il kernel
   NETLINK_CB(buf).dst_pid = payload->sender_pid;  // PID a cui  destinato il messaggio
   NETLINK_CB(buf).dst_groups = 0; // Unicast communication
   netlink_unicast(nl_sk, buf, payload->sender_pid, MSG_DONTWAIT);
   printk(KERN_ALERT "  Message '%s' resent to %d process\n", 
	  payload->str,
	  NETLINK_CB(buf).dst_pid);
}


/** _init_module non fa altro che creare il socket NETLINK.
 * La creazione del socket a livello kernel si esegue con `netlink_kernel_create()'.
 * Non si usa netlink_create() la quale viene comunque richiamata da _kernel_create.
 * */
int nl_init_module(void)
{
   nl_sk = netlink_kernel_create(NETLINK_TEST, // TEST: valora arbitrario deciso da noi
				 nl_input);    // Funzione di gestione del socket
   if(nl_sk)
     printk(KERN_INFO "Socket creato: %u\n", (unsigned)nl_sk);
   else
     printk(KERN_ALERT "Socket non creato: %u\n", (unsigned)nl_sk);
   
   return 0;
}


/** _cleanup_module deve assolutamente eliminare il socket NETLINK.
 * NOTA: cosa succede sa il socket  aperto (x es in user-space) e scarico il 
 * modulo?
 * */
void nl_cleanup_module(void)
{
   printk(KERN_INFO "Close NETLINK socket %u...\n", (unsigned)nl_sk);
   sock_release(nl_sk->sk_socket);
}

module_init(nl_init_module);
module_exit(nl_cleanup_module);
